#include <translation/TranslatorAddOn.h>
#include <translation/TranslatorFormats.h>
#include <translation/BitmapStream.h>
#include <support/ByteOrder.h>
#include <FindDirectory.h>
#include <PopUpMenu.h>
#include <MenuItem.h>
#include <MenuField.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <Path.h>
#include <Locker.h>
#include "zlib.h"
#include "Layer.h"

size_t GetLine (BPositionIO *source, char *line, size_t max);
void Merge (BBitmap *a, Layer *b, BRect update, bool doselect, bool preserve_alpha);
uchar clipchar (int x);

#define BUF_SIZE 4096
#define BECASSO_CODE 'SBec'
#define MAX_LAYERS 32
#define BLEND_USES_SHIFTS

struct bt_settings
{
	BPoint  origin;
	int32	z_compression;// = 6; //Z_BEST_COMPRESSION;	// settable
	bool	settings_touched;
};
BLocker g_settings_lock ("BT Settings Lock");

bt_settings g_settings;

BPoint get_window_origin ();
void set_window_origin (BPoint origin);
void show_settings ();

BPoint get_window_origin ()
{
	BPoint origin;
	g_settings_lock.Lock();
	origin = g_settings.origin;
	g_settings_lock.Unlock();
	return origin;
}

void set_window_origin (BPoint origin)
{
	g_settings_lock.Lock();
	g_settings.origin = origin;
	g_settings.settings_touched = true;
	g_settings_lock.Unlock();
}

void show_settings ()
{
	g_settings_lock.Lock();
	printf ("Window Position = (%g, %g)\n", g_settings.origin.x, g_settings.origin.y);
	printf ("Compression Level [0..9] = %ld\n", g_settings.z_compression);
	g_settings_lock.Unlock();
}

class PrefsLoader
{
public:
			 PrefsLoader ()
			 {
			 	// printf ("PrefsLoader... "); fflush (stdout);
			 	g_settings_lock.Lock();
			 	// defaults:
				g_settings.origin = B_ORIGIN;
				g_settings.z_compression = 6;
				g_settings.settings_touched = false;
				
				BPath path;
				if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
				{
					path.SetTo ("/tmp");
				}
				path.Append ("BecassoTranslator_Settings");
				FILE *fp = fopen (path.Path(), "r");
				if (fp)
				{
					char line[1024];
					char name[32];
					char *ptr;
					while (true)
					{
						line[0] = 0;
						fgets (line, 1024, fp);
						if (!line[0])
							break;
						ptr = line;
						while (isspace (*ptr))
							ptr++;
						if (*ptr == '#' || !*ptr)	// Comment line
							continue;
						if (sscanf (ptr, "%31[a-zA-Z_0-9] =", name) != 1)
						{
							fprintf (stderr, "Strange BecassoTranslator settings line: %s\n", name);
						}
						else
						{
							if (!strcmp (name, "window_origin"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if (sscanf (ptr, "%f,%f", &g_settings.origin.x, &g_settings.origin.y) != 2)
									fprintf (stderr, "Illegal window origin in Becasso Translator settings: %s\n", ptr);
							}
							else if (!strcmp (name, "compression_level"))
							{
								while (*ptr != '=')
									ptr++;
								ptr++;
								if ((sscanf (ptr, "%ld", &g_settings.z_compression) != 1) || (g_settings.z_compression < 0) || (g_settings.z_compression > 9))
								{
									fprintf (stderr, "Illegal compression level in Becasso Translator settings: %s\n", ptr);
									g_settings.z_compression = 6;
								}
							}
							else
							{
								fprintf (stderr, "Unknown BecassoTranslator setting: %s\n", name);
							}
						}
					}
					fclose (fp);
				}
				g_settings_lock.Unlock();
				// printf ("Done.\n");
			 }
			 
			~PrefsLoader ()
			 {
			 	// printf ("~PrefsLoader... "); fflush (stdout);
			 	g_settings_lock.Lock();
			 	if (g_settings.settings_touched)
			 	{
			 		BPath path;
			 		if (find_directory (B_USER_SETTINGS_DIRECTORY, &path))
			 		{
			 			path.SetTo ("/tmp");
			 		}
			 		path.Append ("BecassoTranslator_Settings");
			 		FILE *fp = fopen (path.Path(), "w");
			 		if (fp)
			 		{
			 			fprintf (fp, "# BecassoTranslator settings - Sum Software (http://www.sumware.demon.nl)\n");
			 			fprintf (fp, "window_origin = %g,%g\n", g_settings.origin.x, g_settings.origin.y);
			 			fprintf (fp, "compression_level = %ld\n", g_settings.z_compression);
			 			fclose (fp);
			 		}
			 	}
			 	g_settings_lock.Unlock();
			 	// printf ("Done.\n");
			 }
};

PrefsLoader g_prefsloader;

status_t Identify (BPositionIO *inSource, const translation_format */*inFormat*/, BMessage */*ioExtension*/, translator_info *outInfo, uint32 outType)
{
	// printf ("BecassoTranslator: Identify\n");
	inSource->Seek (0L, SEEK_SET);
	char header[32];
	size_t number_read = 0;
	number_read = inSource->Read (header, 20L);
	header[number_read] = 0;
	if (number_read)
	{
		if (!strcmp (header, "# Becasso image file"))
		// We have a Becasso file in *inSource!
		{
			if (outType == BECASSO_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= BECASSO_CODE;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;	// Lossless compression used
				outInfo->capability	= 0.1;	// Let's be modest for now.
				strcpy (outInfo->MIME, "image/x-becasso");
				strcpy (outInfo->name, "Becasso Image");
				// printf ("Returning OK -> can translate to Becasso\n");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == BECASSO_CODE || outType == B_TRANSLATOR_BITMAP || !outType)
			{
				outInfo->type		= B_TRANSLATOR_BITMAP;
				outInfo->group		= B_TRANSLATOR_BITMAP;
				outInfo->quality	= 0.9;
				outInfo->capability	= 0.1;
				strcpy (outInfo->MIME, "image/x-be-bitmap");
				strcpy (outInfo->name, "B_TRANSLATOR_BITMAP");
				// printf ("Returning OK -> can translate from Becasso\n");
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

size_t GetLine (BPositionIO *source, char *line, size_t max)
{
	size_t numread = 0;
	while (numread < max && source->Read (line + numread, 1) && line[numread++] != '\n')
		;
	return numread;
}

inline uchar clipchar (int x)
{
	return ((x < 0) ? 0 : ((x > 255) ? 255 : x));
}

void Merge (BBitmap *a, Layer *b, BRect update, bool doselect, bool preserve_alpha)
{
//	clock_t start, end;
//	start = clock();
	doselect = doselect;
	preserve_alpha = preserve_alpha;
//	ulong h = update.IntegerHeight() + 1;	// Was: fCanvasFrame!
	ulong w = update.IntegerWidth() + 1;
//	ulong sbbpr = a->BytesPerRow();
	ulong sbpr = b->BytesPerRow()/4;
	ulong rt = ulong (update.top);
	ulong rb = ulong (update.bottom);
	ulong rl = ulong (update.left);
	ulong rr = ulong (update.right);
	ulong dw = rr - rl + 1;
	ulong ddiff = (w - dw);
	ulong sdiff = sbpr - dw;
//	ulong brl = rl;
	uint32 *src = (uint32 *) b->Bits() + rt*sbpr + rl - 1;
	uint32 *dest = (uint32 *) a->Bits() + (rt*w + rl) - 1;
	int ga = b->getGlobalAlpha();
	switch (b->getMode())
	{
	case 0:
		for (ulong y = rt; y <= rb; y++)
		{
			for (ulong x = rl; x <= rr; x++)
			{
#if defined (__POWERPC__)
				register uint32 srcpixel = *(++src);
				register uint32 destpixel = *(++dest);
				register int sa = (srcpixel & 0xFF)*ga/255;
				register int da = 255 - sa;
				if (da == 0)		// Fully opaque pixel
				{
					*dest = srcpixel;
				}
				else if (da == 255)	// Fully transparent pixel
				{
				}
				else
				{
#	if !defined (BLEND_USES_SHIFTS)
					*dest = ((((destpixel & 0xFF000000)/255*da + (srcpixel & 0xFF000000)/255*sa)) & 0xFF000000) |
							((((destpixel & 0x00FF0000)*da + (srcpixel & 0x00FF0000)*sa)/255) & 0x00FF0000) |
							((((destpixel & 0x0000FF00)*da + (srcpixel & 0x0000FF00)*sa)/255) & 0x0000FF00) |\
							  // (max_c (sa, destpixel & 0xFF));
							    (clipchar (sa + int (destpixel & 0xFF)));
#	else
					*dest = (((((destpixel & 0xFF000000)>>8)*da + ((srcpixel & 0xFF000000)>>8)*sa)) & 0xFF000000) |
							 ((((destpixel & 0x00FF0000)*da + (srcpixel & 0x00FF0000)*sa)>>8) & 0x00FF0000) |
							 ((((destpixel & 0x0000FF00)*da + (srcpixel & 0x0000FF00)*sa)>>8) & 0x0000FF00) |\
							    (clipchar (sa + int (destpixel & 0xFF)));
#	endif
				}
#else
				register uint32 srcpixel = *(++src);
				register uint32 destpixel = *(++dest);
				register int sa = (srcpixel >> 24)*ga/255;
				register int da = 255 - sa;
				if (da == 0)		// Fully opaque pixel
				{
					*dest = srcpixel;
				}
				else if (da == 255)	// Fully transparent pixel
				{
				}
				else
				{
#	if !defined (BLEND_USES_SHIFTS)
					*dest = ((((destpixel & 0x00FF0000)*da + (srcpixel & 0x00FF0000)*sa)/255) & 0x00FF0000) |
							((((destpixel & 0x0000FF00)*da + (srcpixel & 0x0000FF00)*sa)/255) & 0x0000FF00) |
							((((destpixel & 0x000000FF)*da + (srcpixel & 0x000000FF)*sa)/255) & 0x000000FF) |
							    (clipchar (sa + int (destpixel >> 24)) << 24);
#	else
					*dest = ((((destpixel & 0x00FF0000)*da + (srcpixel & 0x00FF0000)*sa)>>8) & 0x00FF0000) |
							((((destpixel & 0x0000FF00)*da + (srcpixel & 0x0000FF00)*sa)>>8) & 0x0000FF00) |
							((((destpixel & 0x000000FF)*da + (srcpixel & 0x000000FF)*sa)>>8) & 0x000000FF) |
							    (clipchar (sa + int (destpixel >> 24)) << 24);
#	endif
				}
#endif
			}
			src += sdiff;
			dest += ddiff;
		}
		break;
	case 1:
		for (ulong y = rt; y <= rb; y++)
		{
			for (ulong x = rl; x <= rr; x++)
			{
#if defined (__POWERPC__)
				register uint32 srcpixel = *(++src);
				register uint32 destpixel = *(++dest);
				register int sa = (srcpixel & 0xFF)*ga/255;
				register int da = destpixel & 0xFF;
				register int tsa = 65025 - sa*255;
				register int tda = 65025 - da*255;
				if (sa == 0)		// Fully transparent pixel
				{
				}
				else if (sa == 255)	// Fully opaque pixel
				{
					*dest = (((((srcpixel >> 24)       )*(tda + da*((destpixel >> 24)       ))/65025) << 24) & 0xFF000000) |
							(((((srcpixel >> 16) & 0xFF)*(tda + da*((destpixel >> 16) & 0xFF))/65025) << 16) & 0x00FF0000) |
							(((((srcpixel >>  8) & 0xFF)*(tda + da*((destpixel >>  8) & 0xFF))/65025) <<  8) & 0x0000FF00) |
							    (clipchar (sa + int (destpixel & 0xFF)));
				}
				else
				{
					*dest = (((tsa + sa*((srcpixel >> 24)       ))*(tda + da*((destpixel >> 24)       ))/16581375 << 24) & 0xFF000000) |
							(((tsa + sa*((srcpixel >> 16) & 0xFF))*(tda + da*((destpixel >> 16) & 0xFF))/16581375 << 16) & 0x00FF0000) |
							(((tsa + sa*((srcpixel >>  8) & 0xFF))*(tda + da*((destpixel >>  8) & 0xFF))/16581375 <<  8) & 0x0000FF00) |
						      (clipchar (sa + int (destpixel & 0xFF)));
				}
#else
				register uint32 srcpixel = *(++src);
				register uint32 destpixel = *(++dest);
				register int sa = (srcpixel >> 24)*ga/255;
				register int da = destpixel >> 24;
				register int tsa = 65025 - sa*255;
				register int tda = 65025 - da*255;
				if (sa == 0)		// Fully transparent pixel
				{
				}
				else if (sa == 255)	// Fully opaque pixel
				{
					*dest = (((((srcpixel >> 16) & 0xFF)*(tda + da*((destpixel >> 16) & 0xFF))/65025) << 16) & 0x00FF0000) |
							(((((srcpixel >>  8) & 0xFF)*(tda + da*((destpixel >>  8) & 0xFF))/65025) <<  8) & 0x0000FF00) |
							(((((srcpixel      ) & 0xFF)*(tda + da*((destpixel      ) & 0xFF))/65025)      ) & 0x000000FF) |
							    (clipchar (sa + int (destpixel >> 24)) << 24);
				}
				else
				{
					*dest = (((tsa + sa*((srcpixel >> 16) & 0xFF))*(tda + da*((destpixel >> 16) & 0xFF))/16581375 << 16) & 0x00FF0000) |
							(((tsa + sa*((srcpixel >>  8) & 0xFF))*(tda + da*((destpixel >>  8) & 0xFF))/16581375 <<  8) & 0x0000FF00) |
							(((tsa + sa*((srcpixel      ) & 0xFF))*(tda + da*((destpixel      ) & 0xFF))/16581375      ) & 0x000000FF) |
							  (clipchar (sa + int (destpixel >> 24)) << 24);
				}
#endif
			}
			src += sdiff;
			dest += ddiff;
		}
		break;
	default:
		fprintf (stderr, "Merge:  Unknown Drawing Mode\n");
		break;
	}
//	end = clock();
//	printf ("Merge took %d ms\n", end - start);
}

status_t Translate (BPositionIO *inSource, const translator_info */*inInfo*/, BMessage */*ioExtension*/, uint32 outType, BPositionIO *outDestination)
{
	inSource->Seek (0L, SEEK_SET);
	char header[32];
	ssize_t number_read = 0;
	number_read = inSource->Read (header, 20L);
	header[number_read] = 0;
	if (number_read)
	{
		if (!strcmp (header, "# Becasso image file"))
		// We have a Becasso file in *inSource!
		{
			if (outType == BECASSO_CODE)
			// "translate" it to Becasso format...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// decode the Becasso image to a bitmap
			{
				//printf ("Translate Becasso image to regular bitmap\n");
				Layer *layer[MAX_LAYERS];
				BRect lFrame;

				// Do it here...
				inSource->Seek (0L, SEEK_SET);
//				bool done = false;
				char line[81];
				do
				{
					GetLine (inSource, line, 80);
				}	while (line[0] == '#');
				char *endp = line;
				float vers = strtod (endp, &endp);
				int fNumLayers = strtol (endp, &endp, 10);
				//printf ("#layers: %i, %s\n", fNumLayers, endp);
				/*int fCurrentLayer = */strtol (endp, &endp, 10);
				switch (int (vers))
				{
				case 0:
					for (int i = 0; i < fNumLayers; i++)
					{
						GetLine (inSource, line, 80);
						line[strlen(line) - 1] = 0;
						endp = line;
						lFrame.left = strtod (endp, &endp);
						lFrame.top = strtod (endp, &endp);
						lFrame.right = strtod (endp, &endp);
						lFrame.bottom = strtod (endp, &endp);
						/*int m = */strtol (endp, &endp, 10);
						bool h = strtol (endp, &endp, 10);
						layer[i] = new Layer (lFrame, endp);
						layer[i]->Hide (h);
						layer[i]->setMode (0);	// Note: DM_BLEND has moved to 0, and <1.1 didn't have any other ops anyway.
						inSource->Read (layer[i]->Bits(), layer[i]->BitsLength());
					}
					break;
				case 1:                        
				{
					uchar *tmpzbuf = NULL;
					z_streamp zstream = new z_stream;
					zstream->zalloc = Z_NULL;
					zstream->zfree = Z_NULL;
					zstream->opaque = Z_NULL;
					if (inflateInit (zstream) != Z_OK)
					{
						fprintf (stderr, "Oops!  Problems with inflateInit()...\n");
					}
					for (int i = 0; i < fNumLayers; i++)
					{
						GetLine (inSource, line, 80);
						line[strlen(line) - 1] = 0;
						endp = line;
						lFrame.left = strtod (endp, &endp);
						lFrame.top = strtod (endp, &endp);
						lFrame.right = strtod (endp, &endp);
						lFrame.bottom = strtod (endp, &endp);
						int m = strtol (endp, &endp, 10);
						bool h = strtol (endp, &endp, 10);
						uchar ga = strtol (endp, &endp, 10);
						/*int am =*/ strtol (endp, &endp, 10);
						layer[i] = new Layer (lFrame, endp);
						layer[i]->Hide (h);
						if (vers == 1.0)
							layer[i]->setMode (0);	// Note: DM_BLEND has moved to 0, and <1.1 didn't have any other ops anyway.
						else
							layer[i]->setMode (m);
						layer[i]->setGlobalAlpha (ga);
			//			layer[i]->setAlphaMap (NULL);
					}
					for (int i = 0; i < fNumLayers; i++)
					{
						// printf ("Decompressing layer %d\n", i);
						if (!tmpzbuf)
						{
							tmpzbuf = new uchar [layer[i]->BitsLength()];
							zstream->avail_in = 0;
						}
						zstream->next_in = tmpzbuf;
						zstream->avail_in += inSource->Read (tmpzbuf, layer[i]->BitsLength() - zstream->avail_in);
						zstream->next_out = (uchar *) layer[i]->Bits();
						zstream->avail_out = layer[i]->BitsLength();
			//			printf ("avail_in = %li, avail_out = %li\n", zstream->avail_in, zstream->avail_out);
						if (inflate (zstream, Z_FINISH) != Z_STREAM_END)
						{
							fprintf (stderr, "Oops!  Layer couldn't be decompressed completely...\n");
						}
						memmove (tmpzbuf, zstream->next_in, zstream->avail_in);
						inflateReset (zstream);
					}
					if (inflateEnd (zstream) != Z_OK)
					{
						fprintf (stderr, "Oops!  Temporary zlib buffer couldn't be deallocated\n");
					}
					delete [] tmpzbuf;
					delete zstream;
					break;
				}
				case 2:                        
				{
					uchar *tmpzbuf = NULL;
					z_streamp zstream = new z_stream;
					zstream->zalloc = Z_NULL;
					zstream->zfree = Z_NULL;
					zstream->opaque = Z_NULL;
					if (inflateInit (zstream) != Z_OK)
					{
						fprintf (stderr, "Oops!  Problems with inflateInit()...\n");
					}
					for (int i = 0; i < fNumLayers; i++)
					{
						GetLine (inSource, line, 80);
						line[strlen(line) - 1] = 0;
						endp = line;
						BRect lFrame;
						lFrame.left = strtod (endp, &endp);
						lFrame.top = strtod (endp, &endp);
						lFrame.right = strtod (endp, &endp);
						lFrame.bottom = strtod (endp, &endp);
						int m = strtol (endp, &endp, 10);
						bool h = strtol (endp, &endp, 10);
						uchar ga = strtol (endp, &endp, 10);
						if (vers > 1.1)
						{
							/* int am = */strtol (endp, &endp, 10);
							// printf ("%salpha map\n", am ? "" : "no ");
						}
						layer[i] = new Layer (lFrame, endp);
						layer[i]->Hide (h);
						if (vers == 1.0)
							layer[i]->setMode (0);	// Note: DM_BLEND has moved to 0, and <1.1 didn't have any other ops anyway.
						else
							layer[i]->setMode (m);
						layer[i]->setGlobalAlpha (ga);
						layer[i]->setAlphaMap (NULL);
					}
					for (int i = 0; i < fNumLayers; i++)
					{
						if (!tmpzbuf)
						{
							tmpzbuf = new uchar [layer[i]->BitsLength()];
							zstream->avail_in = 0;
						}
						zstream->next_in = tmpzbuf;
						zstream->avail_in += inSource->Read (tmpzbuf, layer[i]->BitsLength() - zstream->avail_in);
						zstream->next_out = (uchar *) layer[i]->Bits();
						zstream->avail_out = layer[i]->BitsLength();
			//			printf ("avail_in = %li, avail_out = %li\n", zstream->avail_in, zstream->avail_out);
						if (inflate (zstream, Z_FINISH) != Z_STREAM_END)
						{
							fprintf (stderr, "Oops!  Layer couldn't be decompressed completely...\n");
						}
						memmove (tmpzbuf, zstream->next_in, zstream->avail_in);
						inflateReset (zstream);
					}
					if (inflateEnd (zstream) != Z_OK)
					{
						fprintf (stderr, "Oops!  Temporary zlib buffer couldn't be deallocated\n");
					}
					delete [] tmpzbuf;
					delete zstream;
					
					break;
				}
				default:	// This should never happen
					fprintf (stderr, "Hmm, somehow a newer version file got through to the actual loading routine.\n");
					return B_NO_TRANSLATOR;
				}
				
				TranslatorBitmap tbitmap;
				int width, height;
				width = lFrame.IntegerWidth() + 1;
				height = lFrame.IntegerHeight() + 1;
				tbitmap.magic    = B_TRANSLATOR_BITMAP;
				tbitmap.bounds   = lFrame;
				tbitmap.rowBytes = width*4;
				tbitmap.colors   = B_RGB_32_BIT;
				tbitmap.dataSize = height*width*4;
				
//				tbitmap.bounds.PrintToStream();
				swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_HOST_TO_BENDIAN);
				swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_HOST_TO_BENDIAN);
				
				// OK, we've got the layers.  Now blend them into a BBitmap.
				BRect update = lFrame;
				update.OffsetTo (B_ORIGIN);
				BBitmap *bitmap = new BBitmap (update, B_RGB_32_BIT);

//				ulong h = height;
				ulong w = width;
//				ulong sbbpr = bitmap->BytesPerRow();
				ulong sbpr = layer[0]->BytesPerRow();
				ulong rt = ulong (update.top);
				ulong rb = ulong (update.bottom);
				ulong rl = ulong (update.left);
				ulong rr = ulong (update.right);
				ulong dw = rr - rl + 1;
				ulong ddiff = (w - dw);
//				ulong sdiff = sbpr - dw;
				uint32 *src = (uint32 *) layer[0]->Bits();
				src += rt*sbpr + rl;
				uint32 *dest =  (uint32 *) bitmap->Bits() + (rt*w + rl) - 1;
//				ulong brl = rl;
			
				//printf ("Background layer: [%ld, %ld, %ld, %ld]\n", rl, rt, rr, rb);
				// printf ("3"); fflush (stdout);
				#if defined (__POWERPC__)
					uint32 check1 = 0xFFFFFF00;
					uint32 check2 = 0xFFFFFF00;
				#else
					uint32 check1 = 0x00FFFFFF;
					uint32 check2 = 0x00FFFFFF;
				#endif
				for (ulong y = rt; y <= rb; y++)	// First layer: Background.
				{
					for (ulong x = rl; x <= rr; x++)
					{
						if (((x & 8) || (y & 8)) && !((x & 8) && (y & 8)))
						{
							*(++dest) = check1;
						}
						else
						{
							*(++dest) = check2;
						}
					}
					dest += ddiff;
				}
				for (int i = 0; i < fNumLayers; i++)	// Next the layers: Add.
				{
					//printf ("Layer %i\n", i);
					if (!layer[i]->IsHidden())
					{
						Merge (bitmap, layer[i], update, false, true);
					}
				}
				
				// Write out the bitmap header
				int32 err;
				err = outDestination->SetSize (sizeof (TranslatorBitmap) + width*height*4);
				if (err < 0)
				{
					fprintf (stderr, "BecassoTranslator: SetSize returns %ld\n", err);
					return err;
				}
				err = outDestination->Write (&tbitmap, sizeof (TranslatorBitmap));
				if (err < 0)
				{
					fprintf (stderr, "BecassoTranslator: Write (header) returns %ld\n", err);
					return err;
				}
				
				// write out the bitmap data
				err = outDestination->Write (bitmap->Bits(), bitmap->BitsLength());
				if (err < 0)
				{
					fprintf (stderr, "BecassoTranslator: Write (data) returns %ld\n", err);
					return err;
				}
				
				for (int i = 0; i < fNumLayers; i++)
					delete layer[i];
				delete bitmap;
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else if (ntohl (*((int32 *) header)) == B_TRANSLATOR_BITMAP)
		// We have a Translator BBitmap in *inSource!
		{
			if (outType == BECASSO_CODE)
			// encode the bitmap to Becasso...
			{
				// Get info on the bitmap
				TranslatorBitmap tbitmap;
				inSource->Seek (0L, SEEK_SET);
				if (inSource->Read (&tbitmap, sizeof (TranslatorBitmap)) != sizeof (TranslatorBitmap))
				{
					fprintf (stderr, "BecassoTranslator: Couldn't read header from inStream\n");
					return B_ERROR;
				}
				swap_data (B_INT32_TYPE, &tbitmap.magic, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_RECT_TYPE, &tbitmap.bounds, sizeof (BRect), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.rowBytes, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.colors, sizeof (color_space), B_SWAP_BENDIAN_TO_HOST);
				swap_data (B_INT32_TYPE, &tbitmap.dataSize, sizeof (uint32), B_SWAP_BENDIAN_TO_HOST);
				
				BRect bounds = tbitmap.bounds;
				
//				int32 width  = bounds.IntegerWidth() + 1;
//				int32 height = bounds.IntegerHeight() + 1;

				switch (tbitmap.colors)
				{
					case B_RGB32:
					case B_RGBA32:
						break;
					default:
						fprintf (stderr, "BecassoTranslator: Unsupported color_space\n");
						return B_ERROR;
				}
				BBitmap *layer = new BBitmap (bounds, B_RGB_32_BIT);

				// And now read the whole thing in our bitmap...
				// printf ("Read: %d bytes\n", inSource->Read (image, tbitmap.dataSize));
				inSource->Read (layer->Bits(), tbitmap.dataSize);
				
				// Encode it here...
				outDestination->Write ("# Becasso image file\n", 21);
				outDestination->Write ("# Generated by BecassoTranslator\n", 33);
				outDestination->Write ("1.1 1 0 72\n", 11);		// Version 1.1 compatible, 1 layer

				uchar *tmpzbuf = new uchar [uint32 (layer->BitsLength()*1.002) + 12L];
				z_streamp zstream = new z_stream;
				zstream->zalloc = Z_NULL;
				zstream->zfree = Z_NULL;
				zstream->opaque = Z_NULL;
				if (deflateInit (zstream, Z_DEFAULT_COMPRESSION) != Z_OK)
				{
					fprintf (stderr, "zlib error!\n");
				}
				char line [128];
				sprintf (line, "%.0f %.0f %.0f %.0f %i %i %i %i %s\n", 
				layer->Bounds().left,
				layer->Bounds().top,
				layer->Bounds().right,
				layer->Bounds().bottom,
				0, // layer->getMode(),
				0, //layer->IsHidden() ? 1 : 0,
				255, //layer->getGlobalAlpha(),
				0, //layer->getAlphaMap() ? 1 : 0,
				"Image"); //layer->getName());
				outDestination->Write (line, strlen (line));

				//printf ("Compressing layer\n");
				zstream->next_in = (uchar *) layer->Bits();
				zstream->avail_in = layer->BitsLength();
				zstream->next_out = tmpzbuf;
				zstream->avail_out = zstream->avail_in;
				if (deflate (zstream, Z_FINISH) != Z_STREAM_END)
				{
					fprintf (stderr, "Argh, zlib couldn't compress the whole layer!\n");
				}
				outDestination->Write (tmpzbuf, zstream->next_out - tmpzbuf);

				// deflateReset (zstream);
				deflateEnd (zstream);
				delete zstream;
				delete [] tmpzbuf;
				delete layer;

				return B_OK;
			}
			else if (outType == B_TRANSLATOR_BITMAP || !outType)
			// "translate" the bitmap...
			{
				outDestination->Write (header, number_read);
				char buffer[BUF_SIZE];
				do
				{
					number_read = inSource->Read (buffer, BUF_SIZE);
					outDestination->Write (buffer, number_read);
				} while (number_read > 0);
				return B_OK;
			}
			else
				return B_NO_TRANSLATOR;
		}
		else
		// Sorry, we can't handle this thing.
			return B_NO_TRANSLATOR;
	}
	return B_NO_TRANSLATOR;
}

class BTConfigView : public BView
{
public:
			 BTConfigView (BRect rect, const char *name, int32 resizing_mode, int32 flags)
			 : BView (rect, name, resizing_mode, flags)
			 {
				SetViewColor (220, 220, 220, 255);
			 	compressionPU = new BPopUpMenu ("Compression");
				compressionPU->AddItem (new BMenuItem ("0 (none)", new BMessage ('pPUc'), '0'));
				for (int32 i = 1; i < 9; i++)
				{
					char label[16];	// quite safe
					sprintf (label, "%ld", i);
					compressionPU->AddItem (new BMenuItem (label, new BMessage ('pPUc'), label[0]));
				}
				compressionPU->AddItem (new BMenuItem ("9 (highest)", new BMessage ('pPUc'), '9'));
			 	compressionMF = new BMenuField (BRect (8, 70, 192, 80), "compressionMF", "Compression:", compressionPU);
			 	compressionMF->SetDivider (80);
			 	g_settings_lock.Lock();
			 	compressionPU->ItemAt(g_settings.z_compression)->SetMarked (true);
				AddChild (compressionMF);
				g_settings_lock.Unlock();
			 }

virtual		~BTConfigView ()
			 {
			 }

virtual void AllAttached ()
			 {
			 	inherited::AllAttached();
			 	BMessenger messenger (this);
			 	for (int i = 0; i < compressionPU->CountItems(); i++)
			 	{
			 		BMenuItem *item = compressionPU->ItemAt (i);
			 		if (item)
			 			item->SetTarget (messenger);
			 	}
			 }

virtual void Draw (BRect update)
			 {
			 	SetFont (be_bold_font);
			 	SetDrawingMode (B_OP_OVER);
			 	DrawString ("Becasso Translator", BPoint (8, 16));
			 	SetFont (be_plain_font);
			 	char vstring[80];
			 	sprintf (vstring, "v %.2f", (float) translatorVersion/100.0);
			 	DrawString (vstring, BPoint (8, 32));
			 	DrawString ("© 1999, 2000 ∑ Sum Software", BPoint (8, 48));
			 	BView::Draw (update);
			 }

virtual void MessageReceived (BMessage *msg)
			 {
			 	switch (msg->what)
			 	{
				case 'pPUc':
					g_settings_lock.Lock();
					g_settings.z_compression = compressionPU->IndexOf (compressionPU->FindMarked());
					g_settings.settings_touched = true;
					g_settings_lock.Unlock();
					break;
			 		
				default:
					inherited::MessageReceived (msg);
				}
			 }
			 
private:
typedef BView inherited;
BPopUpMenu	*compressionPU;
BMenuField	*compressionMF;
};

status_t MakeConfig (BMessage */*ioExtension*/, BView **outView, BRect *outExtent)
{
	BView *view = new BTConfigView (BRect (0, 0, 200, 100), "BecassoTranslator Settings", B_FOLLOW_ALL, B_WILL_DRAW);
	*outView = view;
	*outExtent = view->Bounds();
	return B_NO_ERROR;
}

char translatorInfo[] = "© ∑ Sum Software <http://www.sumware.demon.nl>";

char translatorName[] = "Becasso Translator";

int32 translatorVersion = 20;

translation_format inputFormats[] = {
	{ BECASSO_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-becasso", "Becasso image format" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};

translation_format outputFormats[] = {
	{ BECASSO_CODE, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-becasso", "Becasso image format" },
	{ B_TRANSLATOR_BITMAP, B_TRANSLATOR_BITMAP, 0.9, 0.1, "image/x-be-bitmap", "Be Bitmap format" },
	{ 0, 0, 0, 0, 0, 0 }
};
